# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import logging
import os
from logging import StreamHandler
from logging.handlers import RotatingFileHandler
from typing import Optional
from datatailr import User
from datatailr.wrapper import dt__Tag


def get_log_level() -> int:
    log_level = os.getenv("DATATAILR_LOG_LEVEL", "INFO").upper()
    if log_level in ["TRACE", "DEBUG"]:
        return logging.DEBUG
    elif log_level == "INFO":
        return logging.INFO
    elif log_level == "WARNING":
        return logging.WARNING
    elif log_level == "ERROR":
        return logging.ERROR
    elif log_level == "CRITICAL":
        return logging.CRITICAL
    else:
        return logging.INFO


tag = dt__Tag()
node_name = tag.get("node_name") or "local"
node_ip = tag.get("node_ip")
job_name = os.getenv("DATATAILR_JOB_NAME", "unknown_job")

try:
    user = User.signed_user().name
except Exception:
    import getpass

    user = getpass.getuser()

LOG_FORMAT = f"%(asctime)s - %(levelname)s - {node_name}:{node_ip} - {user} - {job_name} - %(name)s - [Line %(lineno)d]: %(message)s"


class DatatailrLogger:
    def __init__(
        self,
        name: str,
        log_file: Optional[str] = None,
        log_level: int = get_log_level(),
        log_format: str = LOG_FORMAT,
    ):
        """
        Initialize the DatatailrLogger.

        :param name: Name of the logger.
        :param log_file: Optional file to log messages to.
        :param log_level: Logging level (default: logging.INFO).
        """
        self.logger = logging.getLogger(name)
        self.logger.setLevel(log_level)

        # Stream handler for stdout/stderr
        stream_handler = StreamHandler()
        stream_handler.setLevel(log_level)
        stream_formatter = logging.Formatter(log_format)
        stream_handler.setFormatter(stream_formatter)
        self.logger.addHandler(stream_handler)

        # Optional file handler
        if log_file:
            file_handler = RotatingFileHandler(
                log_file, maxBytes=10 * 1024 * 1024, backupCount=5
            )
            file_handler.setLevel(log_level)
            file_formatter = logging.Formatter(log_format)
            file_handler.setFormatter(file_formatter)
            self.logger.addHandler(file_handler)
        self.enable_opentelemetry()

    def get_logger(self):
        """
        Get the configured logger instance.

        :return: Configured logger.
        """
        return self.logger

    def enable_opentelemetry(self):
        """
        Enable OpenTelemetry integration if available.
        """
        try:
            from opentelemetry.instrumentation.logging import LoggingInstrumentor  # type: ignore

            LoggingInstrumentor().instrument(set_logging_format=True)
        except ImportError:
            pass  # OpenTelemetry is not installed, skip instrumentation



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

from __future__ import annotations

import contextvars
import json
import os
from functools import reduce
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union
import uuid

from datatailr import Image
from datatailr.errors import BatchJobError
from datatailr.logging import DatatailrLogger
from datatailr.scheduler.base import (
    ACL,
    EntryPoint,
    Environment,
    Job,
    JobType,
    Resources,
    User,
)
from datatailr.scheduler.constants import DEFAULT_TASK_CPU, DEFAULT_TASK_MEMORY
from datatailr.scheduler.arguments_cache import ArgumentsCache
from datatailr.scheduler.schedule import Schedule
from datatailr.utils import is_dt_installed

__DAG_CONTEXT__: contextvars.ContextVar = contextvars.ContextVar("dag_context")
__ARGUMENTS_CACHE__ = ArgumentsCache()
logger = DatatailrLogger(os.path.abspath(__file__)).get_logger()


def get_current_manager():
    return __DAG_CONTEXT__.get(None)


class CyclicDependencyError(BatchJobError):
    """
    Exception raised when a cyclic dependency is detected in the batch job dependencies.
    """

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class DuplicateJobNameError(BatchJobError):
    """
    Exception raised when a job with a duplicate name is added to the batch.
    """

    def __init__(self, job_name: str):
        super().__init__(
            f"A job with the name '{job_name}' already exists in the batch."
        )
        self.job_name = job_name


class MissingDagError(BatchJobError):
    """
    Exception raised when a BatchJob is created outside the context of a Batch.
    """

    def __init__(self):
        super().__init__(
            "A BatchJob must be either created within the context of a Batch or a Batch object has to be provided as the dag argument."
        )


class CodePackageMismatchError(BatchJobError):
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class BatchJob:
    """
    Represents a job within a batch job.

    This class can be extended to define specific configurations for each job in the batch.
    """

    def __init__(
        self,
        name: str,
        entrypoint: EntryPoint,
        resources: Optional[Resources] = None,
        dependencies: Sequence[Union[str, BatchJob]] = [],
        dag: Optional[Batch] = get_current_manager(),
        argument_mapping: Dict[str, str] = {},
    ):
        self.name = name
        self.entrypoint = entrypoint
        self.resources = resources
        self.dependencies: set = set(dependencies)
        if dag is None:
            raise MissingDagError()
        self.__id = dag.next_job_id
        self.dag = dag
        self.__args: Dict[str, Any] = {}
        self.dag.__BATCH_JOB_NAMES__[self.name] = self.__id
        self.dependencies = self.translate_dependencies()
        assert all(
            isinstance(dep, int) for dep in self.dependencies
        ), "All dependencies must be integers representing job IDs."
        self.dag.add_job(self)
        self.argument_mapping = argument_mapping or {}

    def __call__(self, *args, **kwds) -> BatchJob:
        """
        Allows the BatchJob instance to be called like a function, returning itself.
        This is useful for chaining or functional-style programming.
        """
        return self

    @property
    def args(self) -> Dict[str, Any]:
        """
        Returns the arguments for the BatchJob instance.
        """
        return self.__args or {}

    @args.setter
    def args(self, args: Dict[str, Any]):
        """
        Sets the arguments for the BatchJob instance.
        """
        if not isinstance(args, dict):
            raise TypeError(f"Expected a dictionary for args, got {type(args)}")
        self.__args = args

    @property
    def id(self) -> int:
        """
        Returns the unique identifier of the BatchJob instance.
        """
        return self.__id

    def alias(self, name: str) -> BatchJob:
        """
        Set an alias for the BatchJob instance.

        :param name: The alias name to set.
        """
        if name in self.dag.__BATCH_JOB_NAMES__:
            raise DuplicateJobNameError(name)
        assert self.dag.__BATCH_JOB_NAMES__.pop(self.name) == self.__id
        self.dag.__BATCH_JOB_NAMES__[name] = self.__id
        self.name = name
        return self

    def set_resources(
        self,
        resources: Optional[Resources] = None,
        memory: Optional[str] = None,
        cpu: Optional[float] = None,
    ) -> BatchJob:
        """
        Set the resources for the BatchJob instance.

        :param resources: The Resources instance to set.
        """
        if resources is not None:
            if not isinstance(resources, Resources):
                raise TypeError(f"Expected Resources instance, got {type(resources)}")
        else:
            resources = Resources(
                memory=memory or DEFAULT_TASK_MEMORY, cpu=cpu or DEFAULT_TASK_CPU
            )
        self.resources = resources
        return self

    def __repr__(self):
        return (
            f"BatchJob(name={self.name}, entrypoint={self.entrypoint}, "
            f"resources={self.resources}) (id={self.__id})"
        )

    def __getstate__(self) -> object:
        state = self.__dict__.copy()
        state.pop("dag", None)
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)

    def to_dict(self):
        """
        Convert the BatchJob instance to a dictionary representation.
        """
        return {
            "display_name": self.name,
            "name": self.name,
            "child_number": self.__id,
            "entrypoint": str(self.entrypoint),
            "memory": self.resources.memory if self.resources else DEFAULT_TASK_MEMORY,
            "cpu": self.resources.cpu if self.resources else DEFAULT_TASK_CPU,
            "depends_on": list(self.dependencies),
        }

    def to_json(self):
        """
        Convert the BatchJob instance to a JSON string representation.
        """
        return json.dumps(self.to_dict())

    def translate_dependencies(self) -> Set[int]:
        """
        Translate the dependencies of the BatchJob instance into a format suitable for the batch job.
        """

        def get_dependency_name(dep):
            if isinstance(dep, str):
                return dep
            elif isinstance(dep, BatchJob):
                return dep.name
            else:
                raise TypeError(f"Unsupported dependency type: {type(dep)}")

        return set(
            [
                self.dag.__BATCH_JOB_NAMES__[get_dependency_name(dep)]
                for dep in self.dependencies
            ]
        )

    def __add_dependency__(self, other):
        self.dependencies.add(other.__id)
        arg_name = self.argument_mapping.get(other.name, other.name)
        if arg_name is not None:
            self.__args[arg_name] = other

    def __lshift__(
        self, other: Sequence[BatchJob] | BatchJob
    ) -> Sequence[BatchJob] | BatchJob:
        if isinstance(other, list):
            for task in other:
                self.__add_dependency__(task)
        else:
            self.__add_dependency__(other)
        return other

    def __rshift__(
        self, other: Sequence[BatchJob] | BatchJob
    ) -> Sequence[BatchJob] | BatchJob:
        if isinstance(other, Sequence):
            for task in other:
                task.__add_dependency__(self)
        else:
            other.__add_dependency__(self)
        return other

    def __rrshift__(self, other: Sequence[BatchJob] | BatchJob) -> BatchJob:
        self.__lshift__(other)
        return self

    def __rlshift__(self, other: Sequence[BatchJob] | BatchJob) -> BatchJob:
        self.__rshift__(other)
        return self

    def __hash__(self):
        return self.__id

    def __lt__(self, other: BatchJob) -> bool:
        return self.__id < other.__id

    def run(self):
        """
        Execute the job's entrypoint.
        """
        if isinstance(self.entrypoint, EntryPoint):
            env = {
                "DATATAILR_BATCH_ID": str(self.dag.id),
                "DATATAILR_JOB_ID": str(self.__id),
                "DATATAILR_JOB_NAME": self.name,
            }
            self.entrypoint(env=env)
        else:
            raise TypeError(f"Invalid entrypoint type: {type(self.entrypoint)}")


class Batch(Job):
    """
    Represents a batch job in the scheduler.

    Inherits from Job and is used to define batch jobs with specific configurations.
    """

    def __init__(
        self,
        name: str,
        environment: Optional[Environment] = Environment.DEV,
        schedule: Optional[Schedule] = None,
        image: Optional[Image] = None,
        run_as: Optional[Union[str, User]] = None,
        resources: Resources = Resources(memory="100m", cpu=1),
        acl: Optional[ACL] = None,
        local_run: bool = False,
        python_requirements: str = "",
        build_script_pre: str = "",
        build_script_post: str = "",
    ):
        super().__init__(
            environment=environment,
            name=name,
            image=image,
            run_as=run_as,
            resources=resources,
            acl=acl,
            python_requirements=python_requirements,
            build_script_pre=build_script_pre,
            build_script_post=build_script_post,
            type=JobType.BATCH,
        )
        self.__jobs: List[BatchJob] = []
        self._auto_run = False
        self.__next_job_id = -1
        self.__BATCH_JOB_NAMES__: Dict[str, int] = {}
        self.__local_run = local_run
        self.__schedule = schedule

    @property
    def next_job_id(self):
        """
        Returns a generator for the next job ID in the batch.
        """
        self.__next_job_id += 1
        return self.__next_job_id

    def add_job(self, job: BatchJob):
        """
        Adds a job to the batch job.

        :param job: The BatchJob instance to add.
        """
        if not isinstance(job, BatchJob):
            raise TypeError(
                f"Only BatchJob instances can be added to a Batch. Got {type(job)} instead."
            )
        if self.get_job_by_name(job.name) is not None:
            raise DuplicateJobNameError(job.name)
        # Use the batch level resource values as defaults for jobs
        job.resources = job.resources or self.resources
        image_path_to_repo = self.image.path_to_repo
        image_path_to_module = self.image.path_to_module
        package_path_to_repo = job.entrypoint.path_to_repo
        package_path_to_module = job.entrypoint.path_to_module

        if image_path_to_repo is None:
            self.image.path_to_repo = package_path_to_repo
        elif package_path_to_repo != image_path_to_repo:
            raise CodePackageMismatchError(
                f"Function {job.entrypoint.function_name} is defined in a different package root: "
                f"{package_path_to_repo} != {image_path_to_repo}"
            )
        if image_path_to_module is None:
            self.image.path_to_module = package_path_to_module
        elif package_path_to_module != image_path_to_module:
            raise CodePackageMismatchError(
                f"Function {job.entrypoint.function_name} is defined in a different module: "
                f"{package_path_to_module} != {image_path_to_module}"
            )
        self.__jobs.append(job)

    def is_job_in(self, job: BatchJob) -> bool:
        return job in self.__jobs

    def get_job_by_name(self, job_name: str) -> Optional[BatchJob]:
        return next((job for job in self.__jobs if job.name == job_name), None)

    def to_dict(self):
        """
        Convert the Batch instance to a dictionary representation.
        """
        batch_dict = super().to_dict()
        batch_dict["jobs"] = [job.to_dict() for job in self.__jobs]
        batch_dict["schedule"] = str(self.__schedule) if self.__schedule else None
        return batch_dict

    def to_json(self):
        """
        Convert the Batch instance to a JSON string representation.
        """
        return json.dumps(self.to_dict(), indent=4)

    def __repr__(self):
        return (
            f"Batch(name={self.name}, environment={self.environment}, "
            f"run_as={self.run_as}, resources={self.resources}, "
            f"acl={self.acl}, {len(self.__jobs)} jobs)"
        )

    def set_autorun(self, auto_run):
        self._auto_run = auto_run

    def __enter__(self):
        self._token = __DAG_CONTEXT__.set(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        __DAG_CONTEXT__.reset(self._token)
        if self._auto_run:
            self.run()

    def __topological_sort__(self):
        jobs = {
            hash(job): set([hash(dep) for dep in job.dependencies])
            for job in self.__jobs
        }

        for k, v in jobs.items():
            v.discard(k)  # ignore self dependencies
        extra_items_in_deps = reduce(set.union, jobs.values()) - set(jobs.keys())
        jobs.update({item: set() for item in extra_items_in_deps})
        while True:
            ordered = set(item for item, dep in jobs.items() if not dep)
            if not ordered:
                break
            yield sorted(ordered)
            jobs = {
                item: (dep - ordered)
                for item, dep in jobs.items()
                if item not in ordered
            }
        if jobs:
            raise CyclicDependencyError(
                "A cyclic dependency exists amongst {}".format(jobs)
            )

    def get_schedule_args(self) -> Dict[str, Any]:
        if isinstance(self.__schedule, Schedule):
            args = {
                "at_minutes": self.__schedule.at_minutes,
                "every_minute": self.__schedule.every_minute,
                "at_hours": self.__schedule.at_hours,
                "every_hour": self.__schedule.every_hour,
                "weekdays": self.__schedule.weekdays,
                "day_of_month": self.__schedule.day_of_month,
                "in_month": self.__schedule.in_month,
                "every_month": self.__schedule.every_month,
                "timezone": self.__schedule.timezone,
                "run_after_job_uuid": self.__schedule.run_after_job_uuid,
                "run_after_job_name": self.__schedule.run_after_job_name,
                "run_after_job_condition": self.__schedule.run_after_job_condition,
            }
            args = {key: value for key, value in args.items() if value is not None}
            for key, value in args.items():
                if isinstance(value, list):
                    args[key] = ",".join(map(str, value))
            return args
        return {}

    def run(self) -> Tuple[bool, str]:
        def arg_name(arg: Union[BatchJob, str]) -> str:
            return arg.name if isinstance(arg, BatchJob) else arg

        def merged(dst: dict[str, str], src: dict[str, str]) -> dict[str, str]:
            # copy so we don't mutate the original mapping
            out = dict(dst)
            seen_vals = set(out.values())
            for k, v in src.items():
                if v not in seen_vals:
                    out[k] = v
                    seen_vals.add(v)
            return out

        args = {
            j.name: merged(
                j.argument_mapping, {k: arg_name(v) for k, v in j.args.items()}
            )
            for j in self.__jobs
        }

        __ARGUMENTS_CACHE__.add_arguments(self.id, args)
        if not self.__local_run and is_dt_installed():
            return super().run()
        else:
            os.environ["DATATAILR_BATCH_RUN_ID"] = uuid.uuid4().hex[:8]
            for step in self.__topological_sort__():
                for job_id in step:
                    job = self.__jobs[job_id]
                    logger.info(
                        f"Batch {self.name}, running job '{job.name}' in environment '{self.environment}' as '{self.run_as}'"
                    )
                    job.run()
            from datatailr.scheduler.batch_decorator import __FUNCTIONS_CREATED_IN_DAG__

            __FUNCTIONS_CREATED_IN_DAG__.clear()
            return True, ""



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

from __future__ import annotations
from typing import Optional

from datatailr.wrapper import dt__User

# Datatailr User API Client
__client__ = dt__User()


class User:
    """
    Representing a Datatailr User.

    This class provides methods to interact with the Datatailr User API.
    It allows you to create, update, delete, and manage users within the Datatailr platform.

    Attributes:
        first_name (str): The first name of the user.
        last_name (str): The last name of the user.
        name (str): The username of the user.
        email (str): The email address of the user.
        user_id (int): The unique identifier for the user.
        primary_group_id (int): The primary group of the user.
        is_system_user (bool): Indicates if the user is a system user.

    Static Methods:
        signed_user() -> Optional['User']:
            Retrieve the currently signed-in user, if available.
        add(name: str, first_name: str = None, last_name: str = None, email: str = None, password: str = None, primary_group_id: int = None, is_system_user: bool = False) -> 'User':
            Create a new user with the specified username, first name, last name, and email.
        get(name: str) -> 'User':
            Retrieve a user by their username.
        exists(name: str) -> bool:
            Check if a user exists by their username.
        ls() -> list:
            List all users available in the Datatailr platform.
        remove(name: str) -> None:
            Remove a user by their username.

    Instance Methods:
        verify() -> None:
            Refresh the user information from the Datatailr API.
    """

    def __init__(self, name):
        self.__name = name
        self.__first_name = None
        self.__last_name = None
        self.__email = None
        self.__user_id = None
        self.__primary_group_id = None
        self.__is_system_user = False
        self.__expiry__ = None
        self.__signature__ = None

        self.__refresh__()

    def __repr__(self):
        return (
            f"User(name={self.name}, first_name={self.first_name}, "
            f"last_name={self.last_name}, email={self.email}, "
            f"user_id={self.user_id}, primary_group_id={self.primary_group_id}, "
            f"is_system_user={self.is_system_user})"
        )

    def __str__(self):
        return f"<User: {self.name} | {self.user_id}>"

    def __eq__(self, other):
        if not isinstance(other, User):
            return NotImplemented
        return (
            self.user_id == other.user_id
            and self.name == other.name
            and self.email == other.email
            and self.primary_group_id == other.primary_group_id
            and self.is_system_user == other.is_system_user
            and self.first_name == other.first_name
            and self.last_name == other.last_name
        )

    def __refresh__(self):
        if not self.name:
            raise ValueError("Name is not set. Cannot refresh user.")
        user = __client__.get(self.name)
        if user:
            self.__name = user["name"]
            self.__first_name = user["first_name"]
            self.__last_name = user["last_name"]
            self.__email = user.get("email")
            self.__user_id = user["user_id"]
            self.__primary_group_id = user["primary_group_id"]
            self.__is_system_user = user["is_system"]

    @property
    def first_name(self):
        return self.__first_name

    @property
    def last_name(self):
        return self.__last_name

    @property
    def name(self):
        return self.__name

    @property
    def email(self):
        return self.__email

    @property
    def user_id(self):
        return self.__user_id

    @property
    def primary_group_id(self):
        return self.__primary_group_id

    @property
    def primary_group(self):
        if self.__primary_group_id is not None:
            from datatailr.group import Group

            return Group.get(self.__primary_group_id)
        return None

    @property
    def is_system_user(self):
        return self.__is_system_user

    @staticmethod
    def get(name: str) -> User:
        return User(name)

    @staticmethod
    def signed_user() -> User:
        user_signature_and_expiry = __client__.signed_user()
        if user_signature_and_expiry:
            user = User(name=user_signature_and_expiry["name"])
            user.__expiry__ = user_signature_and_expiry["expiry"]
            user.__signature__ = user_signature_and_expiry["signature"]
            return user

        raise PermissionError(
            "No signed user found. Please ensure you are signed in to Datatailr."
        )

    @staticmethod
    def add(
        name: str,
        first_name: str,
        last_name: str,
        email: str,
        password: str,
        primary_group: str,
        is_system_user: bool = False,
    ) -> Optional["User"]:
        if is_system_user:
            if password is not None:
                raise Warning(
                    "Password is not required for system users. It will be ignored."
                )
            new_user = __client__.add(
                name,
                first_name=first_name,
                last_name=last_name,
                email=email,
                primary_group=primary_group,
                system=is_system_user,
            )
        else:
            new_user = __client__.add(
                name,
                first_name=first_name,
                last_name=last_name,
                email=email,
                password=password,
                primary_group=primary_group,
                system=is_system_user,
            )

        return new_user

    @staticmethod
    def exists(name: str) -> bool:
        return __client__.exists(name)

    @staticmethod
    def ls() -> list:
        users = __client__.ls()
        return [User.get(user["name"]) for user in users]

    @staticmethod
    def remove(name: str) -> None:
        __client__.rm(name)
        return None

    def verify(self) -> None:
        return __client__.verify(self.name, self.__expiry__, self.__signature__)



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import json
import os
import subprocess
from typing import Union

from datatailr.utils import is_dt_installed

API_JSON_PATH: Union[str, None] = None

if is_dt_installed() or os.path.exists("/opt/datatailr/etc/api.json"):
    API_JSON_PATH = os.path.join("/opt", "datatailr", "etc", "api.json")
    CLI_TOOL = "dt"
else:
    # For running local tests, use api.json from the repo and a mock CLI tool
    import sys

    if any("unittest" in arg for arg in sys.argv):
        API_JSON_PATH = os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", "dt", "api.json"
        )
    CLI_TOOL = "echo"


def type_map(arg_type):
    mapping = {
        "String": "str",
        "file": "str",
        "boolean": "bool",
        "int": "int",
        "float": "float",
    }
    return mapping.get(arg_type, "str")


def add_quotes(arg):
    """Add quotes to the argument if it contains spaces or special characters."""
    if isinstance(arg, str) and (" " in arg or '"' in arg or "'" in arg):
        return f'"{arg}"'
    return str(arg)


def make_method(cmd_name, sub_name, sub_def):
    arg_names = sub_def.get("non_options", {}).get("arg_names", [])
    arg_types = sub_def.get("non_options", {}).get("arg_types", [])
    min_args = sub_def.get("non_options", {}).get("min", len(arg_names))
    max_args = sub_def.get("non_options", {}).get("max", len(arg_names))
    options = sub_def.get("options", [])
    option_names = [opt[1] for opt in options]
    description = sub_def.get("description", "")
    help_text = sub_def.get("help", "")
    return_type = "str"

    # Build docstring
    doc = f"""{description}\n\n{help_text}\n"""
    if arg_names:
        doc += "\nArgs:\n"
        for n, t in zip(arg_names, arg_types):
            doc += f"    {n} ({type_map(t)}):\n"
    if options:
        doc += "\nOptions:\n"
        for opt in options:
            short, long, opt_type = opt[:3]
            doc += f"    --{long} ({opt_type})\n"
    doc += f"\nReturns:\n    {return_type}\n"

    def method(self, *args, **kwargs):
        # Accept required args as either positional or keyword arguments
        all_args = list(args)
        used_kwargs = set()
        kwargs = {k.replace("_", "-"): v for k, v in kwargs.items()}
        # Fill missing positional args from kwargs if available
        for i in range(len(all_args), len(arg_names)):
            arg_name = arg_names[i]
            if arg_name in kwargs:
                all_args.append(kwargs[arg_name])
                used_kwargs.add(arg_name)
        # Argument count check
        missing = []
        if len(all_args) < min_args:
            for i in range(len(all_args), min_args):
                if i < len(arg_names):
                    missing.append(
                        f"{arg_names[i]}: {type_map(arg_types[i]) if i < len(arg_types) else 'str'}"
                    )
        if not (min_args <= len(all_args) <= max_args):
            msg = f"{cmd_name}{' ' + sub_name if sub_name else ''} expects between {min_args} and {max_args} arguments, got {len(all_args)}."
            if missing:
                msg += f" Missing: {', '.join(missing)}"
            raise TypeError(msg)
        # Argument type check
        for i, (arg, expected_type) in enumerate(zip(all_args, arg_types)):
            py_type = eval(type_map(expected_type))
            if not isinstance(arg, py_type):
                raise TypeError(
                    f"Argument '{arg_names[i]}' must be of type {py_type.__name__}, got {type(arg).__name__}"
                )
        # Disallow unexpected kwargs (only allow options and used arg-names)
        allowed_kwargs = set(option_names) | set(arg_names)
        unexpected_kwargs = set(kwargs.keys()) - allowed_kwargs
        if unexpected_kwargs:
            raise TypeError(
                f"{cmd_name}{' ' + sub_name if sub_name else ''} got unexpected keyword arguments: {', '.join(unexpected_kwargs)}. Expected: {', '.join(allowed_kwargs)}"
            )
        cmd = [self.cli_tool, cmd_name]
        if sub_name:
            cmd.append(sub_name)
        for arg in all_args:
            cmd.append(add_quotes(str(arg)))
        # Only pass option kwargs (not those used for required args)
        if "json" in allowed_kwargs:
            kwargs["json"] = True  # Ensure JSON output if 'json' is an option
        for k, v in kwargs.items():
            if k in option_names:
                # Find the option definition
                opt_def = next((opt for opt in options if opt[1] == k), None)
                if opt_def and opt_def[2] == "no_argument":
                    if v:
                        cmd.append(f"--{k}")
                    # If v is False, omit it (do nothing)
                else:
                    # For options that require a value
                    cmd.append(f"--{k}={v}")
        result = subprocess.run(cmd, capture_output=True, text=True)
        if result.returncode != 0:
            raise RuntimeError(f"Command failed: {' '.join(cmd)}\n{result.stderr}")
        try:
            return json.loads(result.stdout) if result.stdout else None
        except json.JSONDecodeError:
            return result.stdout.strip()

    method.__doc__ = doc
    return method


def create_class(cmd_name, command):
    class_name = cmd_name.capitalize()
    sub_commands = command.get("sub_commands", {})
    # Create a new class with only cli_tool attribute
    cls = type(class_name, (object,), {"cli_tool": CLI_TOOL})
    if sub_commands:
        for sub_name, sub_def in sub_commands.items():
            sub_name = (
                sub_name  # Convert hyphens to underscores for Python method names
            )
            method = make_method(cmd_name, sub_name, sub_def)
            setattr(cls, sub_name.replace("-", "_"), method)
    else:
        method = make_method(cmd_name, None, command)
        setattr(cls, cmd_name, method)
    return cls


# Load API JSON and create classes at import time
if API_JSON_PATH and os.path.exists(API_JSON_PATH):
    with open(API_JSON_PATH, "r") as f:
        api = json.load(f)

    for cmd_name, command in api.items():
        cls = create_class(cmd_name, command)
        globals()["dt__" + cmd_name.capitalize()] = cls


class mock_cli_tool:
    """
    Mock CLI tool for local usage of the DataTailr platform.
    This function simulates the CLI tool behavior by returning a predefined response.
    """

    # make it possible to use any function, even if it doesn't eist in the class
    def mock_method(self, name, *args, **kwargs):
        def fun(*args, **kwargs):
            return {"mocked": name, "args": args, "kwargs": kwargs}

        return fun

    def __getattr__(self, name):
        return self.mock_method(name)


dt__User = globals().get("dt__User", mock_cli_tool)
dt__Group = globals().get("dt__Group", mock_cli_tool)
dt__Job = globals().get("dt__Job", mock_cli_tool)
dt__Blob = globals().get("dt__Blob", mock_cli_tool)
dt__Dns = globals().get("dt__Dns", mock_cli_tool)
dt__System = globals().get("dt__System", mock_cli_tool)
dt__Sms = globals().get("dt__Sms", mock_cli_tool)
dt__Email = globals().get("dt__Email", mock_cli_tool)
dt__Kv = globals().get("dt__Kv", mock_cli_tool)
dt__Log = globals().get("dt__Log", mock_cli_tool)
dt__Node = globals().get("dt__Node", mock_cli_tool)
dt__Tag = globals().get("dt__Tag", mock_cli_tool)
dt__Registry = globals().get("dt__Registry", mock_cli_tool)
dt__Service = globals().get("dt__Service", mock_cli_tool)
dt__Settings = globals().get("dt__Settings", mock_cli_tool)



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

from datatailr.wrapper import dt__System, mock_cli_tool
from datatailr.group import Group
from datatailr.user import User
from datatailr.acl import ACL
from datatailr.blob import Blob
from datatailr.build import Image
from datatailr.utils import Environment, is_dt_installed
from datatailr.version import __version__

system = dt__System()
if isinstance(system, mock_cli_tool):
    __provider__ = "not installed"
else:
    __provider__ = system.provider()

__all__ = [
    "ACL",
    "Blob",
    "Environment",
    "Group",
    "Image",
    "User",
    "__version__",
    "__provider__",
    "is_dt_installed",
]



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

r"""
Datatailr Scheduler Module
==========================

The `datatailr.scheduler` module provides a framework for scheduling and managing batch jobs.

The main job types are:
_______________________

- **Batch**: Represents a batch job that can be scheduled and executed.
  The job can include multiple tasks which can be run in parallel or sequentially.
- **Service**: Represents a service job that runs continuously.
- **App**: Represents a web app or a dashboard, which can be built using one of the supported frameworks,
  such as `Streamlit <https://streamlit.io/>`_, `Dash <https://dash.plotly.com/>`_, or `Panel <https://panel.holoviz.org/>`_.
- **Excel**: Represents an Excel add-in.
"""

from datatailr.errors import BatchJobError
from datatailr.scheduler.base import (
    EntryPoint,
    Environment,
    Job,
    JobType,
    Resources,
    set_allow_unsafe_scheduling,
)
from datatailr.scheduler.batch import Batch, BatchJob, DuplicateJobNameError
from datatailr.scheduler.batch_decorator import batch_decorator as batch_job
from datatailr.scheduler.schedule import Schedule

__all__ = [
    "Job",
    "JobType",
    "Environment",
    "Resources",
    "EntryPoint",
    "Batch",
    "BatchJob",
    "batch_job",
    "BatchJobError",
    "DuplicateJobNameError",
    "set_allow_unsafe_scheduling",
    "Schedule",
]



#!/usr/bin/env python3

# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************


# The purpose of this script is to be the entrypoint for all jobs running on datatailr.
# The main functions of the script are:
#     1. Create a linux user and group for the job.
#     2. Set the environment variables for the job.
#     3. Run the job in a separate process, as the newly created user and pass all relevant environment variables.
# There are muliple environment variables which are required for the job to run.
# Some of them are necessary for the setup stage, which is executed directly in this script as the linux root user.
# Others are passed to the job script, which is executed in a separate process with only the users' privileges and not as a root user.
#
# Setup environment variables:
#     DATATAILR_USER - the user under which the job will run.
#     DATATAILR_GROUP - the group under which the job will run.
#     DATATAILR_UID - the user ID of the user as it is defined in the system.
#     DATATAILR_GID - the group ID of the group as it is defined in the system.
#     DATATAILR_JOB_TYPE - the type of job to run. (batch\service\app\excel\IDE)
# Job environment variables (not all are always relevant, depending on the job type):
#     DATATAILR_BATCH_RUN_ID - the unique identifier for the batch run.
#     DATATAILR_BATCH_ID - the unique identifier for the batch.
#     DATATAILR_JOB_ID - the unique identifier for the job.


import os
import sys
from typing import Tuple
from datatailr.logging import DatatailrLogger


logger = DatatailrLogger(os.path.abspath(__file__)).get_logger()


def get_env_var(name: str, default: str | None = None) -> str:
    """
    Get an environment variable.
    If the variable is not set, raise an error.
    """
    if name not in os.environ:
        if default is not None:
            return default
        logger.error(f"Environment variable '{name}' is not set.")
        raise ValueError(f"Environment variable '{name}' is not set.")
    return os.environ[name]


def create_user_and_group() -> Tuple[str, str]:
    """
    Create a user and group for the job.
    The user and group names are taken from the environment variables DATATAILR_USER and DATATAILR_GROUP.
    The group and user are created with the same uid and gid as passed in the environment variables DATATAILR_UID and DATATAILR_GID.
    If the user or group already exists, do nothing.
    """
    user = get_env_var("DATATAILR_USER")
    group = get_env_var("DATATAILR_GROUP")
    uid = get_env_var("DATATAILR_UID")
    gid = get_env_var("DATATAILR_GID")

    # Create group if it does not exist
    os.system(f"getent group {group} || groupadd {group} -g {gid} -o")

    # Create user if it does not exist
    os.system(
        f"getent passwd {user} || useradd -g {group} -s /bin/bash -m {user} -u {uid} -o"
    )
    return user, group


def run_command_as_user(command: str, user: str, env_vars: dict):
    """
    Run a command as a specific user with the given environment variables.
    """
    env_vars.update({"PATH": get_env_var("PATH")})
    env_vars.update({"PYTHONPATH": get_env_var("PYTHONPATH", "")})
    env_vars_str = " ".join(f"{key}='{value}'" for key, value in env_vars.items())
    full_command = f"sudo -u {user} {env_vars_str} {command}"
    logger.debug(f"Running command: {full_command}")
    status = os.system(full_command)
    if status != 0:
        logger.error(f"Command failed with status {status / 256:g}")
        sys.exit(1)


def main():
    user, _ = create_user_and_group()
    job_type = get_env_var("DATATAILR_JOB_TYPE")

    job_name = get_env_var("DATATAILR_JOB_NAME")
    job_id = get_env_var("DATATAILR_JOB_ID")

    if job_type == "batch":
        run_id = get_env_var("DATATAILR_BATCH_RUN_ID")
        batch_id = get_env_var("DATATAILR_BATCH_ID")
        entrypoint = get_env_var("DATATAILR_BATCH_ENTRYPOINT")
        env = {
            "DATATAILR_BATCH_RUN_ID": run_id,
            "DATATAILR_BATCH_ID": batch_id,
            "DATATAILR_JOB_ID": job_id,
            "DATATAILR_BATCH_ENTRYPOINT": entrypoint,
        }
        run_command_as_user("datatailr_run_batch", user, env)
    elif job_type == "service":
        port = get_env_var("DATATAILR_SERVICE_PORT")
        entrypoint = get_env_var("DATATAILR_ENTRYPOINT")
        env = {
            "DATATAILR_JOB_NAME": job_name,
            "DATATAILR_JOB_ID": job_id,
            "DATATAILR_ENTRYPOINT": entrypoint,
            "DATATAILR_SERVICE_PORT": port,
        }
        run_command_as_user("datatailr_run_service", user, env)
    elif job_type == "app":
        entrypoint = get_env_var("DATATAILR_ENTRYPOINT")
        env = {
            "DATATAILR_JOB_NAME": job_name,
            "DATATAILR_JOB_ID": job_id,
            "DATATAILR_ENTRYPOINT": entrypoint,
        }
        run_command_as_user("datatailr_run_app", user, env)
    elif job_type == "excel":
        host = get_env_var("DATATAILR_HOST")
        entrypoint = get_env_var("DATATAILR_ENTRYPOINT")
        env = {
            "DATATAILR_JOB_NAME": job_name,
            "DATATAILR_JOB_ID": job_id,
            "DATATAILR_ENTRYPOINT": entrypoint,
            "DATATAILR_HOST": host,
        }
        run_command_as_user("datatailr_run_excel", user, env)
    elif job_type == "IDE":
        pass
    else:
        raise ValueError(f"Unknown job type: {job_type}")


if __name__ == "__main__":
    try:
        logger.debug("Starting job execution...")
        main()
        logger.debug("Job executed successfully.")
    except Exception as e:
        logger.error(f"Error during job execution: {e}")
        raise



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import shutil
from enum import Enum
import subprocess
from typing import Tuple


class Environment(Enum):
    """
    Enum representing different environments for DataTailr jobs.
    """

    DEV = "dev"
    PRE = "pre"
    PROD = "prod"

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"Environment.{self.name}('{self.value}')"


def is_dt_installed():
    """
    Check if DataTailr is installed by looking for the 'dt' command in the system PATH.
    """
    return shutil.which("dt") is not None


def run_shell_command(command: str) -> Tuple[str, int]:
    """
    Run a shell command.

    This function executes a shell command and returns the output.

    Args:
        command (str): The shell command to execute.

    Returns:
        str: The output of the executed command.
    """
    result = subprocess.run(command, shell=True, capture_output=True, text=True)
    if result.returncode != 0:
        raise RuntimeError(f"Command '{command}' failed with error: {result.stderr}")
    return result.stdout.strip(), result.returncode
